#!/usr/bin/perl

###############################################################################
## Copyright (c) 2006 Xilinx, Inc.
## This design is confidential and proprietary of Xilinx, Inc.
## All Rights Reserved
###############################################################################
##   ____  ____
##  /   /\/   /
## /___/  \  /    Vendor: Xilinx
## \   \   \/     Version: 1.0
##  \   \         Filename: mig20_ucf_rev.pl
##  /   /         Date Last Modified:
## /___/   /\     Date Created: 10/15/07
## \   \  /  \
##  \___\/\___\
##
##Device: Virtex-5
##Purpose:
##  Perl script to uprev UCF file generated by MIG 1.7x and earlier for MIG
##  Virtex-5 DDR2 design for use with MIG 2.0 Virtex-5 DDR2 design. This
##  script reads both the customer UCF and corrsponding Virtex-5 PARTGEN
##  generated package file to determine correct values for the corresponding
##  UCF and HDL constraints required for the MIG 2.0 design:
##    1. Generates HDL code snippet showing correct values for DQS_IO_COL,
##       and DQ_IO_MS parameters
##    2. Generates an output UCF file snippet with the following:
##      a. RLOC_ORIGIN for all DQ read data capture RPMs
##      b. LOC constraints for DQS Gate-related IDELAY, IDDR, and sourcing
##         fabric flop-flop
##  The process followed by this perl script is outlined in Answer Record 29313
##  and XAPP858
##  Notes: 
##Reference:
##  AR 29313, XAPP858
##Revision History:
##    Rev 1.0 - created. RC. 10/15/07
##    Rev 1.1 - internal feedback - added automatic call to PARTGEN. Add
##              error msg w/ usage if wrong # of arguments. Changed MAXDELAY
##              for DQS Gate to output 800ps or 850ps based on freq. RC. 
##              10/18/07
###############################################################################

###############################################################################
# Usage:
#  mig20_ucf_rev.pl <input_UCF> <input PKG> <clock_freq> <output_file_name>
# Arguments:
#  <input_UCF>  = customer (pre-MIG 2.0) Virtex-5 DDR2 UCF
#  <part_num>   = target device part/package number. This is used to 
#                 automatically invoke the ISE utility PARTGEN to generate a 
#                 package file for the target part.This # includes the
#                 entire P/N except for the speed grade. E.g. this argument
#                 would for an LX330T in a 1738-pin package is:
#                 'xc5vlx330tff1738'.
#  <clock_freq> = target clock frequency (in MHz)
#  <output_file_name> = base name for output files
# Outputs:
#  <output_file_name>.ucf = UCF showing pinout-specific constraints required
#                           for MIG 2.0 Virtex-5 DDR2 Design
#  <output_file_name>_hdl_params.txt = HDL code snippet showing pinout-specific
#                                      parameters/generics required for
#                                      MIG 2.0 Virtex-5 DDR2 Design
#  <part_num>.pkg = Package file for <part_num> used by this program. This is
#                   only needed for this script - user may delete this file
#                   after running the script.
#  partlist.xct, partlist.xml - files generated by PARTGEN. These are not used
#                              at all by this script. User may delete these
#                              files.
# NOTES:
#  1. Assumes DQ pins follow the naming convention "ddr2_dq[X]" (X = integer)
#  2. Assumes DQS pins: (1) Are differential, and (2) the "N"-side follows the
#     naming convention "ddr2_dqs_n[X]"
#  3. If the above naming conventions are not used, the user can:
#    - Create a temp version of their pre-MIG 2.0 UCF file, and change only
#      the names of the DQ and DQS_N pins to match this convention.
#    - Invoke the script, and generate the updated UCF file. The script only
#      looks at the location ("LOC") constraints for the DQ and DQS_N pins, and
#      ignores everything else.
#    - Copy the contents of this output UCF file to their existing UCF file
###############################################################################

###############################################################################
# Parse the command line arguments
###############################################################################

$usage_msg =
  "Usage:\n".
  " mig20_ucf_rev.pl <input_UCF> <input PKG> <clock_freq> <output_file_name>\n".
  "Arguments:\n".
  " <input_UCF>  = customer (pre-MIG 2.0) Virtex-5 DDR2 UCF\n".
  " <part_num>   = target device part/package number. This is used to\n".
  "                automatically invoke the ISE utility PARTGEN to generate a\n".
  "                package file for the target part.This # includes the\n".
  "                entire P/N except for the speed grade. E.g. this argument\n".
  "                would for an LX330T in a 1738-pin package is:\n".
  "                'xc5vlx330tff1738'.\n".
  " <clock_freq> = target clock frequency (in MHz)\n".
  " <output_file_name> = base name for output files\n";

# Output Usage message if incorrect # of arguments
if (@ARGV != 4) {
  print "ERROR: Incorrect number of arguments!\n";
  print $usage_msg;
  exit(1);
}

$infile_ucf  = $ARGV[0];
$pkg_partnum = $ARGV[1];
$freq        = $ARGV[2];
$outfile     = $ARGV[3];

if (!open(INFILE_UCF, $infile_ucf)) {
  print "ERROR: Can't open input UCF file $infile! ".
    "Exiting program!\n";
  exit(1);
}

if (!open(OUTFILE_UCF, ">$outfile.ucf")) {
  print "ERROR: Can't create output UCF file $outfile.ucf! ".
    "Exiting program! \n";
  exit(1);
}

$outfile_hdl = $outfile."_hdl_params.txt";
if (!open(OUTFILE_HDL, ">$outfile_hdl")) {
  print "ERROR: Can't create HDL param snippet file $outfile_hdl!".
    "Exiting program!\n";
  exit(1);
}

#####################################################################
# Invoking PARTGEN to generate package file for use later in program
#####################################################################

# Delete existing PKG file, just in case PARTGEN does not correctly
# go through, and there is an existing (older, and for some reason)
# incorrect partgen file already in that directory. Probably not likely
# to happen...
unlink("$pkg_partnum.pkg");

print "Invoking PARTGEN to create package file $pkg_partnum.pkg:\n\n";
print "partgen -intstyle xflow -v $pkg_partnum\n";
system("partgen -intstyle xflow -v $pkg_partnum");

# Flag error if PARTGEN fails to produce PKG file
if (!open(INFILE_PKG, "$pkg_partnum.pkg")) {
  print "ERROR: Can't open PARTGEN package file $pkg_partnum.pkg\n";
  print "       Package file not generated. Check if partgen path is set\n";
  print "       and correct part number specified. Exiting program!\n";
  exit(1);
}

print "\n";

###############################################################################
# Read entire PKG file, and store the following information:
#  1. Pin numbers
#  2. IOB coordinates
#  3. Nearest Slice
#  4. Diff pair information (e.g. master or slave)
###############################################################################

print "Processing PKG file: $infile_pkg... ";

$linenum  = 0;
$pinnum   = 0;
$iopinnum = 0;
while (<INFILE_PKG>) {
  chop;
  @entries = split;
  # First line should indicate package part number
  if ($linenum == 0) {
    $partnum = $entries[1];
  }
  # Otherwise, only process lines that have "pin" or "pkgpin"
  # as first entry (although only ones marked "pin" can be used)
  elsif (($entries[0] eq "pin") || ($entries[0] eq "pkgpin")) {
    if ($entries[0] eq "pin") {
      $pinname[$iopinnum]  = $entries[2];
      $nearclbx[$iopinnum] = ($nearclby[$iopinnum] = $entries[6]);
      # Extract numeric part of X-coordinate
      $nearclbx[$iopinnum] =~ s/X//;
      $nearclbx[$iopinnum] =~ s/Y[0-9]+//;
      # Extract numeric part of Y-coordinate
      $nearclby[$iopinnum] =~ s/X[0-9]+//;
      $nearclby[$iopinnum] =~ s/Y//;
      ($diffpair[$iopinnum] = $entries[7]) =~ s/^[0-9]+//;
      # store just IOB coordinates (not IOB name) - might want to add
      # error detection to make sure user only uses IOB, rather than
      # output, or input only pads, for memory bidirectional pins
      ($padname[$iopinnum] = $entries[1]) =~ s/IOB_//;
      $padname[$iopinnum] =~ s/IPAD_//;
      $padname[$iopinnum] =~ s/OPAD_//;
      # Also store which column the IOB is one (0=left,1=mid,2=right)
      # determine this by checking X-coordinate of IOB location
      $iocol[$iopinnum] = ($padname[$iopinnum] =~ m/^X0/) ? 0 :
	(($padname[$iopinnum] =~ m/^X1/) ? 1 : 2);
      ++$iopinnum;
    }
    ++$pinnum;
  }
  ++$linenum;
}
close(INFILE_PKG);
print "Done\n";
print "  Part Number: $partnum\n";
print "  Total pins found: $pinnum\n";
print "  Total I/O pins found: $iopinnum\n";

###############################################################################
# Parse UCF file for DQ, DQS locations
###############################################################################

#####################################################################
# Set DQ, DQS name, numbering format
#####################################################################

$vect_left      = "\\[";     # left side of bus index field
$vect_right     = "\\]";     # right side of bus index field
$vect_zero_fill = 0;         # whether to left pad bux index w/ 0's
$vect_index_len = 2;         # used if zero_fill = 1 (width of bus index)
$dqsn_ending    = "";        # use if dqsn name is, say "ddr2_dqs_00_n"

# Expected signal names - change this if UCF naming convention is different
# from default naming convention
$dq_name   = "ddr2_dq";
$dqsn_name = "ddr2_dqs_n";

$dq_prefix   = $dq_name.$vect_left;
$dqsn_prefix = $dqsn_name.$vect_left;
$dq_suffix   = $vect_right;
$dqsn_suffix = $vect_right.$dqsn_ending;

#####################################################################
# UCF constraint strings
#####################################################################

$rloc_string =
  "INST \"*/gen_dq[sub1].u_iob_dq/gen_stg2_*.u_ff_stg2a_rise\"".
  "\tRLOC_ORIGIN = ";

#####################################################################
# Read UCF file - look for "LOC" constraints for DQ, DQS_N
#####################################################################

print "Processing input UCF file: $infile_ucf... ";

$dq_num  = 0;
$dqs_num = 0;

while (<INFILE_UCF>) {
  chop;
  @entries = split;

  # Isolate pin location - only valid if we find a LOC constraint
  # involving one of the DQ or DQS_N pins. Read in pin location -
  # need to factor out the different exact syntaxes (e.g. space
  # between "LOC" and "=", quotes around the pin name)
  $pinloc = $_;
  # remove everything up to "LOC = "
  $pinloc =~ s/^.*LOC.*=//i;
  # remove everything after closing ";
  $pinloc =~ s/;.*$//;
  # remove any whitespace and quotation marks around pinname
  $pinloc =~ s/[\"|\s]//g;

   # If this is a LOC constraint for a NET
  if ((/^\s*NET/i) && (/LOC\s*=/i)) {
    # Check for DQ
    if ($entries[1] =~ m/$dq_prefix[0-9]+$dq_suffix/) {
      # extract vector number
      $dq_vect = $entries[1];
      $dq_vect =~ s/\"//g;
      $dq_vect =~ s/$dq_prefix//;
      $dq_vect =~ s/$dq_suffix//;
      # read in pin location - need to factor out the different exact
      # syntaxes (e.g. space between "LOC" and "=", quotes around the
      # pin name)
      $pinloc = $_;
      # remove everything up to "LOC = "
      $pinloc =~ s/^.*LOC.*?=//i;
      # remove everything after closing ";
      $pinloc =~ s/;.*$//g;
      # remove any whitespace and quotation marks around pinname
      $pinloc =~ s/[\"|\s]//g;

      # search PKG array for matching pin location
      $pinfound = 0;
      for ($i = 0; $i < $iopinnum; $i++) {
	# if matching pin location found:
	#  1. calculate RLOC_ORIGIN based on I/O column
	#  2. store the value in an array for output later
	if ($pinloc eq $pinname[$i]) {
	  # value of RLOC_ORIGIN depends on: (1) I/O column, and
	  # (2) nearest CLB field
	  $dq_vect = int($dq_vect);
	  $dq_rlocx[$dq_vect] = ($nearclbx[$i] == 0) ?
	    $nearclbx[$i] : $nearclbx[$i] - 4;
	  # Make an adjustment based on if the X-coordinate is
	  # is odd, if so, then add 1 - this is the same thing
	  # as always using the corresponding Master X-coordinate 
	  # for a Slave IOB
	  if (($dq_rlocx[$dq_vect]%2) == 1) {
	    ++$dq_rlocx[$dq_vect];
	  }
	  $dq_rlocy[$dq_vect] = $nearclby[$i];
	  # also save whether master or slave I/O
	  $dq_ms[$dq_vect] = ($diffpair[$i] eq "M") ? 1 : 0;
	  ++$dq_num;
	  $pinfound = 1;
	  last;
	}
      }
      if (!$pinfound) {
	print "\nERROR: Pin \"$pinloc\" is not a valid DQ I/O pin on line $..".
	  " Exiting program!\n";
	exit(1);
      }
    }
    # Check for DQS_N
    elsif ($entries[1] =~ m/$dqsn_prefix[0-9]+$dqsn_suffix/) {
      # determine vector number
      $dqsn_vect = $entries[1];
      $dqsn_vect =~ s/\"//g;
      $dqsn_vect =~ s/$dqsn_prefix//;
      $dqsn_vect =~ s/$dqsn_suffix//;
      # read in pin location - need to factor out the different exact
      # syntaxes (e.g. space between "LOC" and "=", quotes around the
      # pin name)
      $pinloc = $_;
      # remove everything up to "LOC = "
      $pinloc =~ s/^.*LOC.*?=//i;
      # remove everything after closing ";
      $pinloc =~ s/;.*$//;
      # remove any whitespace and quotation marks around pinname
      $pinloc =~ s/[\"|\s]//g;
      # search PKG array for matching pin location
      $pinfound = 0;
      for ($i = 0; $i < $iopinnum; $i++) {
	# if matching pin location found:
	#  1. Calculate/store DQS Gate IDELAY/IDDR location
	#  2. Calculate/store sourcing fabric flop location
	if ($pinloc eq $pinname[$i]) {
	  $dqsn_vect = int($dqsn_vect);
	  $dqsn_iddr[$dqsn_vect]  = $padname[$i];
	  $dqsn_iocol[$dqsn_vect] = $iocol[$i];
	  $dqsn_flopx[$dqsn_vect] = $nearclbx[$i];
	  $dqsn_flopy[$dqsn_vect] = $nearclby[$i];
	  ++$dqs_num;
	  $pinfound = 1;
	  last;
	}
      }
      if (!$pinfound) {
	print "\nERROR: Pin \"$pinloc\" is not a valid DQS I/O pin on line $..".
	  " Exiting program!\n";
	exit(1);
      }
    }
  }
}
close(INFILE_UCF);
print "Done\n";
print "  DQ (data) I/O found: $dq_num\n";
print "  DQS (strobe) I/O found: $dqs_num\n";

###############################################################################
# Write to output UCF file
###############################################################################

#####################################################################
# Define constant UCF constraints
#####################################################################

$now_string = localtime();

$ucf_header =
  "############################################################################\n".
  "##\n".
  "##  Xilinx, Inc. 2006-2007            www.xilinx.com\n".
  "##  $now_string\n".
  "##  Generated by MIG20_UCF_REV.PL:\n".
  "##   Script to generate additional UCF constraints needed by MIG 2.0 V5 DDR2\n".
  "##\n".
  "############################################################################\n".
  "##  File name :   $outfile.ucf\n".
  "##  Details :     Constraints file\n".
  "##                    FPGA family:       virtex5\n".
  "##                    FPGA:              $partnum\n".
  "##                    Frequency:         $freq MHz\n".
  "############################################################################\n".
  "\n".
  "############################################################################\n".
  "# Usage:\n".
  "#   Cut and paste output of this file, and place at end of existing\n".
  "#   (pre-MIG 2.0) UCF file for Virtex-5 DDR2 design\n".
  "#   NOTES:\n".
  "#    1. It may be necessary to change some of the net names in this UCF\n".
  "#       output file to match user's hierarchy, and/or post-synthesized\n".
  "#       netlist\n".
  "#    2. See Answer Record 29313 for the procedure to generate the UCF and\n".
  "#       HDL constraints needed for the MIG 2.0 Virtex-5 DDR2 design - this\n".
  "#       is the procedure followed by this script\n".
  "############################################################################\n".
  "\n";

$ucf_multi_cyc_paths =
  "###############################################################################\n".
  "# Define multicycle paths - these paths may take longer because additional\n".
  "# time allowed for logic to settle in calibration/initialization FSM\n".
  "###############################################################################\n".
  "NET \"clk0\"    TNM = FFS \"TNM_CLK0\";\n".
  "NET \"clk90\"   TNM = FFS \"TNM_CLK90\";\n".
  "# MUX Select for either rising/falling CLK0 for 2nd stage read capture\n".
  "INST \"*/u_phy_calib_0/gen_rd_data_sel*.u_ff_rd_data_sel\" TNM = \"TNM_RD_DATA_SEL\";\n".
  "TIMESPEC \"TS_MC_RD_DATA_SEL\" = FROM \"TNM_RD_DATA_SEL\" TO \"TNM_CLK0\"\n".
  "  \"TS_SYS_CLK\" * 4;\n".
  "# MUX select for read data - optional delay on data to account for byte skews\n".
  "INST \"*/u_usr_rd_0/gen_rden_sel_mux*.u_ff_rden_sel_mux\" TNM = \"TNM_RDEN_SEL_MUX\";\n".
  "TIMESPEC \"TS_MC_RDEN_SEL_MUX\" = FROM \"TNM_RDEN_SEL_MUX\" TO \"TNM_CLK0\"\n".
  "  \"TS_SYS_CLK\" * 4;\n".
  "# Calibration/Initialization complete status flag (for PHY logic only)\n".
  "INST \"*/u_phy_init_0/u_ff_phy_init_data_sel\" TNM = \"TNM_PHY_INIT_DATA_SEL\";\n".
  "TIMESPEC \"TS_MC_PHY_INIT_DATA_SEL_0\" = FROM \"TNM_PHY_INIT_DATA_SEL\" TO\n".
  "  \"TNM_CLK0\" \"TS_SYS_CLK\" * 4;\n".
  "TIMESPEC \"TS_MC_PHY_INIT_DATA_SEL_90\" = FROM \"TNM_PHY_INIT_DATA_SEL\" TO\n".
  "  \"TNM_CLK90\" \"TS_SYS_CLK\" * 4;\n".
  "# Select (address) bits for SRL32 shift registers used in stage3/stage4\n".
  "# calibration\n".
  "INST \"*/u_phy_calib_0/gen_gate_dly*.u_ff_gate_dly\" TNM = \"TNM_GATE_DLY\";\n".
  "TIMESPEC \"TS_MC_GATE_DLY\" = FROM \"TNM_GATE_DLY\" TO \"TNM_CLK0\" \"TS_SYS_CLK\" * 4;\n".
  "INST \"*/u_phy_calib_0/gen_rden_dly*.u_ff_rden_dly\" TNM = \"TNM_RDEN_DLY\";\n".
  "TIMESPEC \"TS_MC_RDEN_DLY\" = FROM \"TNM_RDEN_DLY\" TO \"TNM_CLK0\" \"TS_SYS_CLK\" * 4;\n".
  "INST \"*/u_phy_calib_0/gen_cal_rden_dly*.u_ff_cal_rden_dly\"\n".
  "TNM = \"TNM_CAL_RDEN_DLY\";\n".
  "TIMESPEC \"TS_MC_CAL_RDEN_DLY\" = FROM \"TNM_CAL_RDEN_DLY\" TO \"TNM_CLK0\"\n".
  "  \"TS_SYS_CLK\" * 4;\n".
  "\n";

$ucf_dqs_gate_sect1 =
  "###############################################################################\n".
  "# DQS Read Postamble Glitch Squelch circuit related constraints\n".
  "###############################################################################\n".
  "\n".
  "###############################################################################\n".
  "# LOC placement of DQS-squelch related IDDR and IDELAY elements\n".
  "# Each circuit can be located at any of the following locations:\n".
  "#  1. Ununsed \"N\"-side of DQS diff pair I/O\n".
  "#  2. DM data mask (output only, input side is free for use)\n".
  "#  3. Any output-only site\n".
  "###############################################################################\n".
  "\n";

$ucf_dqs_gate_sect2 =
  "###############################################################################\n".
  "# LOC and timing constraints for flop driving DQS CE enable signal\n".
  "# from fabric logic. Even though the absolute delay on this path is\n".
  "# calibrated out (when synchronizing this output to DQS), the delay\n".
  "# should still be kept as low as possible to reduce post-calibration\n".
  "# voltage/temp variations - these are roughly proportional to the\n".
  "# absolute delay of the path\n".
  "###############################################################################\n".
  "\n";

$ucf_dqs_gate_sect3 =
  "# Control for DQS gate - from fabric flop. Prevent \"runaway\" delay -\n".
  "# two parts to this path: (1) from fabric flop to IDELAY, (2) from\n".
  "# IDELAY to asynchronous reset of IDDR that drives the DQ CE's\n".
  "# A single number is used for all speed grades - value based on 333MHz.\n".
  "# This can be relaxed for lower frequencies.\n".
  "NET \"*/u_phy_io_0/en_dqs*\" MAXDELAY = 600 ps;\n".
  "NET \"*/u_phy_io_0/gen_dqs*.u_iob_dqs/en_dqs_sync\" MAXDELAY = %1 ps;\n".
  "\n".
  "###############################################################################\n".
  "# \"Half-cycle\" path constraint from IDDR to CE pin for all DQ IDDR's\n".
  "# for DQS Read Postamble Glitch Squelch circuit\n".
  "###############################################################################\n".
  "\n".
  "# Max delay from output of IDDR to CE input of DQ IDDRs = tRPST + some slack\n".
  "#  where slack account for rise-time of DQS on board. For now assume slack =\n".
  "#  0.400ns (based on initial SPICE simulations, assumes use of ODT), so\n".
  "#  time = 0.4*Tcyc + 0.40ns = 1.6ns \@333MHz\n".
  "INST \"*/gen_dqs[*].u_iob_dqs/u_iddr_dq_ce\" TNM = \"TNM_DQ_CE_IDDR\";\n".
  "INST \"*/gen_dq[*].u_iob_dq/gen_stg2_*.u_iddr_dq\" TNM = \"TNM_DQS_FLOPS\";\n".
  "TIMESPEC \"TS_DQ_CE\" = FROM \"TNM_DQ_CE_IDDR\" TO \"TNM_DQS_FLOPS\" %2 ns;\n".
  "\n".
  "###############################################################################\n".
  "# Location constraints for DQ read-data capture flops in fabric (for 2nd\n".
  "# stage capture)\n".
  "###############################################################################\n".
  "\n";

$ucf_dqs_gate_iddr_string =
  "INST \"*/gen_dqs[%1].u_iob_dqs/u_iddr_dq_ce\"\tLOC = \"ILOGIC_%2\";\n";
$ucf_dqs_gate_idelay_string =
  "INST \"*/gen_dqs[%1].u_iob_dqs/u_iodelay_dq_ce\"\tLOC = \"IODELAY_%2\";\n";
$ucf_dqs_gate_ff_loc_string =
  "INST \"*/u_phy_calib_0/gen_gate[%1].u_en_dqs_ff\"\tLOC = SLICE_X%2Y%3;\n";
$ucf_rloc_origin_string =
  "INST \"*/gen_dq[%1].u_iob_dq/gen_stg2_*.u_ff_stg2a_rise\"\tRLOC_ORIGIN = X%2Y%3;\n";

#####################################################################
# Print first few (fixed) sections of UCF file (header +
# fixed constraints)
#####################################################################

print "Generating UCF output file $outfile.ucf... ";
print OUTFILE_UCF $ucf_header;
print OUTFILE_UCF $ucf_multi_cyc_paths;

#####################################################################
# Print LOC constraints for DQS Gate IDELAY, IDD
#####################################################################

print OUTFILE_UCF $ucf_dqs_gate_sect1;

for ($i = 0; $i < $dqs_num; $i++) {
  $tmp_string = $ucf_dqs_gate_iddr_string;
  $tmp_string =~ s/%1/$i/;
  $tmp_string =~ s/%2/$dqsn_iddr[$i]/;
  print OUTFILE_UCF $tmp_string;
  $tmp_string = $ucf_dqs_gate_idelay_string;
  $tmp_string =~ s/%1/$i/;
  $tmp_string =~ s/%2/$dqsn_iddr[$i]/;
  print OUTFILE_UCF $tmp_string;
}
print OUTFILE_UCF "\n";

#####################################################################
# Print LOC constraints for DQS Gate fabric flop
#####################################################################

print OUTFILE_UCF $ucf_dqs_gate_sect2;

for ($i = 0; $i < $dqs_num; $i++) {
  $tmp_string = $ucf_dqs_gate_ff_loc_string;
  $tmp_string =~ s/%1/$i/;
  $tmp_string =~ s/%2/$dqsn_flopx[$i]/;
  $tmp_string =~ s/%3/$dqsn_flopy[$i]/;
  print OUTFILE_UCF $tmp_string;
}
print OUTFILE_UCF "\n";

#####################################################################
# Print DQS Gate related timing constraints
#####################################################################

$period_ns = (1/$freq) * 1000.00;  # period in ns
$dqs_gate_from_to = (0.4 * $period_ns) + 0.4;
$dqs_gate_from_to_string = sprintf("%.2f",$dqs_gate_from_to);

$tmp_string = $ucf_dqs_gate_sect3;
# MIG outputs either 800ps or 850ps based on speed grade (-1 = 850).
# For now choose which value based on frequency. Either method will
# be fine (the script method is actually more accurate). The tighter
# spec for higher speed grades was based on our timing tests across
# various devices - we found we could meet the tighter constraint on
# on the faster speed grades) - the motivation was to get as much 
# timing slack as was possible.
if ($freq > 267.00) {
  $tmp_string =~ s/%1/800/;
}
else {
  $tmp_string =~ s/%1/850/;
}
$tmp_string =~ s/%2/$dqs_gate_from_to_string/;
print OUTFILE_UCF $tmp_string;

#####################################################################
# Print RLOC_ORIGIN constraints
#####################################################################

for ($i = 0; $i < $dq_num; $i++) {
  $tmp_string = $ucf_rloc_origin_string;
  $tmp_string =~ s/%1/$i/;
  $tmp_string =~ s/%2/$dq_rlocx[$i]/;
  $tmp_string =~ s/%3/$dq_rlocy[$i]/;
  print OUTFILE_UCF $tmp_string;
}

close(OUTFILE_UCF);
print "Done\n";

###############################################################################
# Write to output HDL snippet file
###############################################################################

$params_header =
  "//**************************************************************************\n".
  "//\n".
  "//  Xilinx, Inc. 2006-2007            www.xilinx.com\n".
  "//  $now_string\n".
  "//  Generated by MIG20_UCF_REV.PL:\n".
  "//   Script to generate additional UCF constraints needed by MIG 2.0 V5 DDR2\n".
  "//\n".
  "//**************************************************************************\n".
  "//  File name :   $outfile_hdl\n".
  "//  Details :     Text file showing correct pinout-specific HDL parameters\n".
  "//                to set for design\n".
  "//                    FPGA family:       virtex5\n".
  "//                    FPGA:              $partnum\n".
  "//**************************************************************************\n".
  "\n".
  "//**************************************************************************\n".
  "// Usage:\n".
  "//   This file shows the correct values for the HDL parameters DQS_IO_COL,\n".
  "//   and DQ_IO_MS. Two code snippets are shown, for Verilog and VHDL. The\n".
  "//   user should cut/paste and substitute in the relevant portion (Verilog\n".
  "//   or VHDL) to the top-level MIG model port map.\n".
  "//   NOTES:\n".
  "//    1. See Answer Record 29313 for the procedure to generate the UCF and\n".
  "//       HDL constraints needed for the MIG 2.0 Virtex-5 DDR2 design - this\n".
  "//       is the procedure followed by this script\n".
  "//**************************************************************************\n".
  "\n";

$params_v =
  "//**************************************************************************\n".
  "// VERILOG: Substitute the following code snippet into the top-level MIG\n".
  "//  DDR2 module parameter list\n".
  "//**************************************************************************\n".
  "\n".
  "   parameter DQS_IO_COL = %1, // I/O column location of DQS groups (=0, left; =1 center, =2 right)\n".
  "   parameter DQ_IO_MS = %2, // Master/Slave location of DQ I/O (=0 slave)\n".
  "\n".
  "\n";

$params_vhd =
  "//**************************************************************************\n".
  "// VHDL: Substitute the following code snippet into the top-level MIG\n".
  "//  DDR2 module generic list\n".
  "//**************************************************************************\n".
  "\n".
  "   DQS_IO_COL : bit_vector := %1; -- I/O column location of DQS groups (=0, left; =1 center, =2 right)\n".
  "   DQ_IO_MS : bit_vector := %2; -- Master/Slave location of DQ I/O (=0 slave)\n".
  "\n".
  "\n";

#####################################################################
# Calculate and print output
#####################################################################

print "Generating HDL output file $outfile_hdl...";

#####################################################################
# Calculate DQS_IO_COL (Which I/O column each DQS group is in)
#####################################################################

$dqs_io_col_bits = 2*$dqs_num;
$dqs_io_col_string = "";
for ($i = 0; $i < $dqs_num; $i++) {
  if ($dqsn_iocol[$i] == 0) {
    $curr_io_col_string = "00";
  }
  elsif ($dqsn_iocol[$i] == 1) {
    $curr_io_col_string = "01";
  }
  elsif ($dqsn_iocol[$i] == 2) {
    $curr_io_col_string = "10";
  }
  else {
    print "\nERROR: Illegal I/O column \"$dqsn_iocol[$i]\"".
      " Exiting program!\n";
    exit(1);
  }
  $dqs_io_col_string = $curr_io_col_string.$dqs_io_col_string;
}
# Need to prepend vector width for Verilog, add quotes for VHDL
$dqs_io_col_string_v   = "$dqs_io_col_bits"."'b".$dqs_io_col_string;
$dqs_io_col_string_vhd = "\"".$dqs_io_col_string."\"";

#####################################################################
# Calculate DQ_MS_IO (DQ data bit master/slave ordering)
#####################################################################

$dq_io_ms_bits = $dq_num;
$dq_io_ms_string = "";
for ($i = 0; $i < $dq_num; $i++) {
  $dq_io_ms_string = "$dq_ms[$i]".$dq_io_ms_string;
}
# Need to prepend vector width for Verilog, add quotes for VHDL
$dq_io_ms_string_v   = "$dq_io_ms_bits"."'b".$dq_io_ms_string;
$dq_io_ms_string_vhd = "\"".$dq_io_ms_string."\""; 

#####################################################################
# Print Verilog and VHDL code snippets
#####################################################################

print OUTFILE_HDL $params_header;

$params_v =~ s/%1/$dqs_io_col_string_v/;
$params_v =~ s/%2/$dq_io_ms_string_v/;
$params_v =~ s/%3/$clk_period_string/;

$params_vhd =~ s/%1/$dqs_io_col_string_vhd/;
$params_vhd =~ s/%2/$dq_io_ms_string_vhd/;

print OUTFILE_HDL $params_v;
print OUTFILE_HDL $params_vhd;

close(OUTFILE_HDL);
print "Done\n";

###############################################################################

print "Program complete. Have a nice day.\n";
